// ================================================================
// GenomeMixer    (c) 2003, Alex Williams
// All rights reserved.
// Send email correspondence to Robert W. Williams <rwilliam@nb.utmem.edu>
// at the University of Tennesse Health Science Center

// This project was funded jointly by the National Institute of Mental Health, National Institute on Drug Abuse,
// and the National Science Foundation (award P20-MH 62009 to KFM and RW)

// This source code may be used according to the terms of the GNU General Public License (GPL).
// See www.gnu.org for specifics. Essentially, the source may be freely used as long as the source code
// for whatever program the source is also made freely available. If you want to use this source
// in a closed-source program, you should email us for licensing options.

// Disclaimer: This code comes with no warranty at all, and there is no guarantee that it will
// work as claimed in the documentation. Hopefully, it will be useful anyway.

// Note that GenomeMixer makes use of the Trolltech Qt toolkit for its window-managing needs.
// There is very little use of Qt data structures (aside from QString) in the actual processing code,
// but **you will need to have Qt installed in order to compile this code.** Check <www.trolltech.com>
// for details; there should be a free GPL-software-only version of Qt available on the Trolltech
// web site.

// Any redistributions of this software in code or binary form must include the
// above copyright notice, licensing information, disclaimer, and this notice itself.
// Basically, all the comments at the top of this file must be retained.
// ================================================================

/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
**
** If you wish to add, delete or rename functions or slots use
** Qt Designer which will update this file, preserving your code. Create an
** init() function in place of a constructor, and a destroy() function in
** place of a destructor.
*****************************************************************************/

const int numberOfPixmaps = 10 + (26); // 10 digits, 26 capital letters

const int minimalSymbolColumnWidth = 13;
const int minimizedNameColWidth = 8;


void MouseInspector::init() {
	mainMixPtr = NULL;
	
	tableInspection->setNumRows(0);
	
	// Load all the pixmaps...
	for (int i = 0; i < numberOfPixmaps; i++) {
		QString nameString;
		if (i <= 9) {
			nameString = QString::number(i);	    
		} else {
			nameString = QString(QChar(i - 10 + 'A'));	    
		}
		//qDebug("Getting map for: " + nameString);
		pixmaps.push_back(QPixmap::fromMimeSource(nameString + QString(".png")));
	}
	
	femChrIcon = QPixmap::fromMimeSource("female_transparent_venus_small.png");
	maleChrIcon = QPixmap::fromMimeSource("male_transparent_mars_small.png");
}

QPixmap MouseInspector::pixmapFromChar(char x) {
	// Returns a pixmap for 0-9 and A-Z.
	if (x >= '0' && x <= '9') {
		assert(x - '0' < numberOfPixmaps);
		return (pixmaps[x - '0']);
	} else if (x >= 'A' && x <= 'Z') {
		assert(x + 10 - 'A' < numberOfPixmaps);
		return (pixmaps[x + 10 - 'A']);
	} else {
		return NULL;
	}	
}

bool MouseInspector::isUpperCaseOrSingleDigit(char x) {
	return ((x >= '0' && x <= '9') || (x >= 'A' && x <= 'Z')); // (x >= 'a' && x <= 'z') ||
}



void MouseInspector::setUpInspector( GenomeMixer *theNewPtr ) {
	mainMixPtr = theNewPtr;
	loadMouseNames();
	setUpHeaders(mainMixPtr->markerColPtr);
	updateCurrentMouse();
	
	if (mainMixPtr->noRecombDistBases == 0) {
		interfereSettings->setText(QString("Interference was turned off for this run of the simulation."));
	} else {
		interfereSettings->setText(QString("Absolute interference: at least ") + QString::number(double(mainMixPtr->noRecombDistBases / 1000000.0)) + " MBases between recombinations.");
	}
	
	switch (mainMixPtr->recombModel) {
	case kNO_RECOMB:
		recombSettings->setText(QString("Recombination was turned off for this run of the simulation."));
		break;
	case kLIN_RECOMB:
		recombSettings->setText(QString("Recombination setting: Using the linear recombination equation."));
		break;
	}
}


void MouseInspector::loadMouseNames() {   
	listInspection->clear();
	if (mainMixPtr == NULL) {
		textTotalMice->setText("0 individuals");
		return;
	}
	
	for (int i = 0; i < mainMixPtr->numMiceToExport(); i++) {
		// Fill up the list with mouse names
		Mouse *tempMouse = mainMixPtr->mouseToExport(i);
		listInspection->insertItem(QString("ID #") + QString::number(tempMouse->ID()) + QString(": ") + QString(tempMouse->name()));
	}
	
	if (mainMixPtr->numMiceToExport() == 1) {
		textTotalMice->setText("1 individual");
	} else {
		textTotalMice->setText(QString::number(mainMixPtr->numMiceToExport()) + QString(" individuals"));
	} 
}


long MouseInspector::countTotalBreakpoints(int whichMouse) {
	long numBreakpoints = 0; // breakpoints per chromosome = std::max(0, numGeneRanges - 1);
	
	assert(mainMixPtr && whichMouse >= 0 && whichMouse < mainMixPtr->numMiceToExport());
	Mouse *theMouse = mainMixPtr->mouseToExport(whichMouse);
	
	for (int chr = 0; chr < theMouse->numChr(); chr++) {
		numBreakpoints += countNumberOfBreakpointsInChr(whichMouse, chr);
	}
	
	return numBreakpoints;
}

long MouseInspector::countNumberOfBreakpointsInChr(int whichMouse, int chr) {
	assert(mainMixPtr && whichMouse >= 0 && whichMouse < mainMixPtr->numMiceToExport());
	Mouse *theMouse = mainMixPtr->mouseToExport(whichMouse);
	
	if (chr == Chromosome::xChrIndex() && theMouse->isMale()) {
		return (theMouse->getChr(kFEMALE, chr)->numGeneRanges() - 1);
	} else if (chr == Chromosome::yChrIndex()) {
		return (theMouse->getChr(kMALE, chr)->numGeneRanges() - 1);
	} else {
		return (theMouse->getChr(kFEMALE, chr)->numGeneRanges() 
				+ theMouse->getChr(kMALE, chr)->numGeneRanges() - 2);
	}
}


void MouseInspector::setUpHeaders(MarkerCollection *markers) {
	if (markers == NULL) {
		// I guess we don't have any markers...
		tableInspection->setNumRows(0);
		tableInspection->setNumCols(0);
		return; // stop, because we don't have any markers, so we don't have any table data, either
	}
	// Set the size of the table
	tableInspection->setNumRows(markers->mostMarkersOnAnyChr());
	// Set the index along the top, where we indicate which chromosome we're on
	tableInspection->setNumCols(Chromosome::numChr() * 3);
	// Re-number the headers along the top, Chromosome 1...N
	for (int i = 0; i < tableInspection->numCols(); i++) {
		
		if (i % 3 == 0) {
			if (int(i/3) == Chromosome::xChrIndex()) {
				tableInspection->horizontalHeader()->setLabel(i, QIconSet(NULL), 
															  QString("Chr X (") +
															  QString::number( int(i / 3) +1) +
															  QString(")")); // Re-number
			} else if (int(i/3) == Chromosome::yChrIndex()) {
				tableInspection->horizontalHeader()->setLabel(i, QIconSet(NULL), 
															  QString("Chr Y (") +
															  QString::number( int(i / 3) +1) +
															  QString(")")); // Re-number
			} else {
				tableInspection->horizontalHeader()->setLabel(i, QIconSet(NULL), 
															  QString("Chr ") + 
															  QString::number( int(i / 3) +1)); // Re-number
			}
			
		} else if (i % 3 == 1) {
			// Maternal data
			tableInspection->horizontalHeader()->setLabel(i, femChrIcon, QString::null);
		} else {
			// Paternal Data
			tableInspection->horizontalHeader()->setLabel(i, maleChrIcon, QString::null);
		}
		
	}
	
	resizeAllColsToFit();
	tableInspection->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
}



void MouseInspector::updateCurrentMouse(void) {
	int theIndex = listInspection->currentItem();     
	if (theIndex < 0 || theIndex >= mainMixPtr->numMiceToExport()) {
		// Just clear everything... there are no mice to get, or this is an invalid selection
		textMouseInfo->setText("No selection.");
		textMouseLineage->setText("Choose an individual from the list.");
		textMaternal->setText("#---");
		textPaternal->setText("#---");
		if (mainMixPtr->numMiceToExport() <= 0) {
			tableInspection->setNumRows(0);
			tableInspection->setNumCols(0);
		}
		return;
	}
	
	Mouse *theMouse = mainMixPtr->mouseToExport(theIndex);
	assert(theMouse != NULL);
	
	textMouseInfo->setText(QString("Mouse ID #")  + QString::number(theMouse->ID()));// + QString(", ") + QString::number(countTotalBreakpoints(theIndex)) + QString("(?) breakpoints"));
	textMouseLineage->setText(QString("Generation ") + QString::number(theMouse->generation()) + QString(": \"") + QString(theMouse->name()) + QString("\""));
	
	// Set the icon based on the mouse's sex
	if (theMouse->isFemale()) {
		iconSexOfThisMouse->setPixmap( QPixmap::fromMimeSource("female_blue_venus_symbol.png"));
	} else {
		iconSexOfThisMouse->setPixmap( QPixmap::fromMimeSource("male_red_mars_symbol.png"));
	}
	
	// Set the mother's ID
	if (theMouse->motherPtr() != NULL) {
		textMaternal->setText(QString("#") + QString::number(theMouse->motherPtr()->ID()));
	} else {
		textMaternal->setText("n/a (purebred)");
	}
	
	// Set the father's ID
	if (theMouse->fatherPtr() != NULL) {
		textPaternal->setText(QString("#") + QString::number(theMouse->fatherPtr()->ID()));
	} else {
		textPaternal->setText("n/a (purebred)");
	}
	
	// Now fill in the table with all the data
	MarkerCollection *markers = mainMixPtr->markerColPtr;
	
	if (markers == NULL) {
		return;
	}   
	
	for (int chr = 0; chr < Chromosome::numChr(); chr++) {
		for (int row = 0; row < markers->markersInChr(chr); row++) {	    
			long markerBaseLoc = markers->base(chr, row);
			
			tableInspection->setText(row, chr * 3, QString(" ") + QString(markers->name(chr, row))); // <-- set the marker name
			
			// Now fill in the cells that indicate the genotype at that marker...
			
			if (chr == Chromosome::xChrIndex() && theMouse->isMale()) {
				// Male mouse, so it only has one X chromosome (from its mother)
				setMarkerCell(row, chr * 3 + 1, theMouse->geneAtBase(kFEMALE, chr, markerBaseLoc));
				clearMarkerCell(row, chr * 3 + 2);
				
			} else if (chr == Chromosome::yChrIndex()) {
				if (theMouse->isMale()) {
					// Y chromosome (so this must be a male mouse), and it only has a Y from its father
					clearMarkerCell(row, chr * 3 + 1);
					setMarkerCell(row, chr * 3 + 2, theMouse->geneAtBase(kMALE, chr, markerBaseLoc));
				} else {
					// or a female, so no Y chromosome data at all
					clearMarkerCell(row, chr * 3 + 1);
					clearMarkerCell(row, chr * 3 + 2);
				}
			} else {
				setMarkerCell(row, chr * 3 + 1, theMouse->geneAtBase(kFEMALE, chr, markerBaseLoc));
				setMarkerCell(row, chr * 3 + 2, theMouse->geneAtBase(kMALE, chr, markerBaseLoc));				
			}
		}
		
	}
	
}


void MouseInspector::clearMarkerCell(int row, int col) {
	tableInspection->setPixmap(row, col, NULL);
	tableInspection->setText(row, col, QString::null);
}

void MouseInspector::setMarkerCell(int row, int col,  int markerValue) {
	if (markerValue < 0) {
		// Clear the cell
		cerr << "Warning: Programming Error: Called MouseInspector::setMarkerCell with an invalid marker value, the value " << markerValue << "." << endl;
		clearMarkerCell(row, col);
		return;
	} else {
		if (markerValue >= 0 && markerValue < mainMixPtr->numParentalStrains()) {
			string theSymbol = mainMixPtr->getPurebredSymbol(markerValue);
			if (theSymbol.size() == 1 && isUpperCaseOrSingleDigit(theSymbol.c_str()[0]) ) {
				// Only if the symbol is exactly one character, and it is an upper-case letter
				// or a digit, do we have an image for it.
				tableInspection->setPixmap(row, col, pixmapFromChar(theSymbol.c_str()[0]));
			} else {
				// I guess whatever this marker's value is, we don't have a picture for it,
				// so just clear the pixmap and set the text
				tableInspection->setPixmap(row, col, NULL);
				tableInspection->setText(row, col, theSymbol);
			}
		} else {
			tableInspection->setPixmap(row, col, NULL);
			tableInspection->setText(row, col, QString("No symbol for ID#") + QString::number(markerValue));
		}
	}
		
}

void MouseInspector::resizeAllColsToFit() {
	for (int col = 0; col < tableInspection->numCols(); col++) {
		if (col % 3 == 0) {
			tableInspection->adjustColumn(col);
		} else if (col % 3 == 1 || col % 3 == 2) {
			tableInspection->setColumnWidth(col, minimalSymbolColumnWidth);
		}
	}	
	for (int row = 0; row < tableInspection->numRows(); row++) {
		tableInspection->setRowHeight(row, 15);
	}
}

void MouseInspector::resizeHideMarkerNames() {
	for (int col = 0; col < tableInspection->numCols(); col++) {
		if (col % 3 == 0) {
			// Marker name
			tableInspection->setColumnWidth(col, minimizedNameColWidth);
		} else if (col % 3 == 1 || col % 3 == 2) {
			// Marker value/symbol
			tableInspection->setColumnWidth(col, minimalSymbolColumnWidth);
		} 
	}	
	for (int row = 0; row < tableInspection->numRows(); row++) {
		tableInspection->setRowHeight(row, 15);
	}
	
}
